"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.DeliLambdaFactory = void 0;
const events_1 = require("events");
const util_1 = require("util");
const common_utils_1 = require("@fluidframework/common-utils");
const server_services_core_1 = require("@fluidframework/server-services-core");
const protocol_base_1 = require("@fluidframework/protocol-base");
const protocol_definitions_1 = require("@fluidframework/protocol-definitions");
const server_services_client_1 = require("@fluidframework/server-services-client");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const utils_1 = require("../utils");
const lambda_1 = require("./lambda");
const checkpointManager_1 = require("./checkpointManager");
// Epoch should never tick in our current setting. This flag is just for being extra cautious.
// TODO: Remove when everything is up to date.
const FlipTerm = false;
const getDefaultCheckpooint = (epoch) => {
    return {
        clients: undefined,
        durableSequenceNumber: 0,
        epoch,
        expHash1: server_services_client_1.defaultHash,
        logOffset: -1,
        sequenceNumber: 0,
        signalClientConnectionNumber: 0,
        term: 1,
        lastSentMSN: 0,
        nackMessages: undefined,
        successfullyStartedLambdas: [],
    };
};
class DeliLambdaFactory extends events_1.EventEmitter {
    constructor(operationsDbMongoManager, collection, tenantManager, clientManager, forwardProducer, signalProducer, reverseProducer, serviceConfiguration) {
        super();
        this.operationsDbMongoManager = operationsDbMongoManager;
        this.collection = collection;
        this.tenantManager = tenantManager;
        this.clientManager = clientManager;
        this.forwardProducer = forwardProducer;
        this.signalProducer = signalProducer;
        this.reverseProducer = reverseProducer;
        this.serviceConfiguration = serviceConfiguration;
    }
    async create(config, context) {
        var _a, _b, _c, _d, _e, _f, _g;
        const { documentId, tenantId, leaderEpoch } = config;
        const sessionMetric = (0, utils_1.createSessionMetric)(tenantId, documentId, server_services_telemetry_1.LumberEventName.SessionResult, this.serviceConfiguration);
        const sessionStartMetric = (0, utils_1.createSessionMetric)(tenantId, documentId, server_services_telemetry_1.LumberEventName.StartSessionResult, this.serviceConfiguration);
        const messageMetaData = {
            documentId,
            tenantId,
        };
        let gitManager;
        let document;
        try {
            const tenant = await this.tenantManager.getTenant(tenantId, documentId);
            gitManager = tenant.gitManager;
            // Lookup the last sequence number stored
            // TODO - is this storage specific to the orderer in place? Or can I generalize the output context?
            document = await this.collection.findOne({ documentId, tenantId });
        }
        catch (error) {
            const errMsg = "Deli lambda creation failed";
            (_a = context.log) === null || _a === void 0 ? void 0 : _a.error(`${errMsg}. Exception: ${(0, util_1.inspect)(error)}`, { messageMetaData });
            this.logSessionFailureMetrics(sessionMetric, sessionStartMetric, errMsg);
            throw error;
        }
        // Check if the document was deleted prior.
        if (!(0, utils_1.isDocumentValid)(document)) {
            // (Old, from tanviraumi:) Temporary guard against failure until we figure out what causing this to trigger.
            // Document sessions can be joined (via Alfred) after a document is functionally deleted.
            (_b = context.log) === null || _b === void 0 ? void 0 : _b.error(`Received attempt to connect to a missing/deleted document.`, { messageMetaData });
            return new utils_1.NoOpLambda(context);
        }
        if (!(0, utils_1.isDocumentSessionValid)(document, this.serviceConfiguration)) {
            // Session for this document is either nonexistent or exists in a different location.
            const errMsg = `Received attempt to connect to invalid session: ${JSON.stringify(document.session)}`;
            (_c = context.log) === null || _c === void 0 ? void 0 : _c.error(errMsg, { messageMetaData });
            if (this.serviceConfiguration.enforceDiscoveryFlow) {
                // This can/will prevent any users from creating a valid session in this location
                // for the liftime of this NoOpLambda. This is not ideal; however, throwing an error
                // to prevent lambda creation would mark the document as corrupted, which is worse.
                return new utils_1.NoOpLambda(context);
            }
        }
        let lastCheckpoint;
        // Restore deli state if not present in the cache. Mongodb casts undefined as null so we are checking
        // both to be safe. Empty sring denotes a cache that was cleared due to a service summary or the document
        // was created within a different tenant.
        if (document.deli === undefined || document.deli === null) {
            (_d = context.log) === null || _d === void 0 ? void 0 : _d.info(`New document. Setting empty deli checkpoint`, { messageMetaData });
            lastCheckpoint = getDefaultCheckpooint(leaderEpoch);
        }
        else {
            if (document.deli === "") {
                (_e = context.log) === null || _e === void 0 ? void 0 : _e.info(`Existing document. Fetching checkpoint from summary`, { messageMetaData });
                const lastCheckpointFromSummary = await this.loadStateFromSummary(tenantId, documentId, gitManager, context.log);
                if (lastCheckpointFromSummary === undefined) {
                    const errMsg = "Could not load state from summary";
                    (_f = context.log) === null || _f === void 0 ? void 0 : _f.error(errMsg, { messageMetaData });
                    this.logSessionFailureMetrics(sessionMetric, sessionStartMetric, errMsg);
                    lastCheckpoint = getDefaultCheckpooint(leaderEpoch);
                }
                else {
                    lastCheckpoint = lastCheckpointFromSummary;
                    // Since the document was originated elsewhere or cache was cleared, logOffset info is irrelavant.
                    // Currently the lambda checkpoints only after updating the logOffset so setting this to lower
                    // is okay. Conceptually this is similar to default checkpoint where logOffset is -1. In this case,
                    // the sequence number is 'n' rather than '0'.
                    lastCheckpoint.logOffset = -1;
                    lastCheckpoint.epoch = leaderEpoch;
                    (_g = context.log) === null || _g === void 0 ? void 0 : _g.info(`Deli checkpoint from summary:
                        ${JSON.stringify(lastCheckpoint)}`, { messageMetaData });
                }
            }
            else {
                lastCheckpoint = JSON.parse(document.deli);
            }
        }
        // For cases such as detached container where the document was generated outside the scope of deli
        // and checkpoint was written manually.
        if (lastCheckpoint.epoch === undefined) {
            lastCheckpoint.epoch = leaderEpoch;
            lastCheckpoint.term = 1;
        }
        const newCheckpoint = FlipTerm ?
            await this.resetCheckpointOnEpochTick(tenantId, documentId, gitManager, context.log, lastCheckpoint, leaderEpoch) :
            lastCheckpoint;
        const checkpointManager = (0, checkpointManager_1.createDeliCheckpointManagerFromCollection)(tenantId, documentId, this.collection);
        // Should the lambda reaize that term has flipped to send a no-op message at the beginning?
        const deliLambda = new lambda_1.DeliLambda(context, tenantId, documentId, newCheckpoint, checkpointManager, this.clientManager, 
        // The producer as well it shouldn't take. Maybe it just gives an output stream?
        this.forwardProducer, this.signalProducer, this.reverseProducer, this.serviceConfiguration, sessionMetric, sessionStartMetric);
        deliLambda.on("close", (closeType) => {
            const handler = async () => {
                var _a;
                if ((closeType === server_services_core_1.LambdaCloseType.ActivityTimeout || closeType === server_services_core_1.LambdaCloseType.Error)) {
                    const query = { documentId, tenantId, session: { $exists: true } };
                    const data = {
                        "session.isSessionAlive": false,
                        "session.isSessionActive": false,
                        "lastAccessTime": Date.now(),
                    };
                    await this.collection.update(query, data, null);
                    (_a = context.log) === null || _a === void 0 ? void 0 : _a.info(`Marked session alive and active as false for closeType: ${JSON.stringify(closeType)}`, { messageMetaData });
                }
            };
            handler().catch((e) => {
                var _a;
                (_a = context.log) === null || _a === void 0 ? void 0 : _a.error(`Failed to handle session alive and active with exception ${e}`, { messageMetaData });
            });
        });
        // Fire-and-forget sessionActive update for session-boot performance.
        // Worst case is that document is allowed to be deleted while active.
        this.collection.update({ documentId, tenantId }, {
            "session.isSessionActive": true,
        }, null).catch((error) => {
            var _a;
            const errMsg = "Deli Lambda failed to mark session as active.";
            (_a = context.log) === null || _a === void 0 ? void 0 : _a.error(`${errMsg}. Exception: ${(0, util_1.inspect)(error)}`, { messageMetaData });
        });
        return deliLambda;
    }
    logSessionFailureMetrics(sessionMetric, sessionStartMetric, errMsg) {
        sessionMetric === null || sessionMetric === void 0 ? void 0 : sessionMetric.error(errMsg);
        sessionStartMetric === null || sessionStartMetric === void 0 ? void 0 : sessionStartMetric.error(errMsg);
    }
    async dispose() {
        var _a;
        const mongoClosedP = this.operationsDbMongoManager.close();
        const forwardProducerClosedP = this.forwardProducer.close();
        const signalProducerClosedP = (_a = this.signalProducer) === null || _a === void 0 ? void 0 : _a.close();
        const reverseProducerClosedP = this.reverseProducer.close();
        await Promise.all([mongoClosedP, forwardProducerClosedP, signalProducerClosedP, reverseProducerClosedP]);
    }
    // Fetches last durable deli state from summary. Returns undefined if not present.
    async loadStateFromSummary(tenantId, documentId, gitManager, logger) {
        const existingRef = await gitManager.getRef(encodeURIComponent(documentId));
        if (existingRef) {
            try {
                const content = await gitManager.getContent(existingRef.object.sha, ".serviceProtocol/deli");
                const summaryCheckpoint = JSON.parse((0, common_utils_1.toUtf8)(content.content, content.encoding));
                return summaryCheckpoint;
            }
            catch (exception) {
                const messageMetaData = {
                    documentId,
                    tenantId,
                };
                logger === null || logger === void 0 ? void 0 : logger.error(`Error fetching deli state from summary`, { messageMetaData });
                logger === null || logger === void 0 ? void 0 : logger.error(JSON.stringify(exception), { messageMetaData });
                return undefined;
            }
        }
    }
    // Check the current epoch with last epoch. If not matched, we need to flip the term.
    // However, we need to store the current term and epoch reliably before we kick off the lambda.
    // Hence we need to create another summary. Logically its an update but in a git sense,
    // its a new commit in the chain.
    // Another aspect is the starting summary. What happens when epoch ticks and we never had a prior summary?
    // For now we are just skipping the step if no prior summary was present.
    // TODO: May be alfred/deli should create a summary at inception?
    async resetCheckpointOnEpochTick(tenantId, documentId, gitManager, logger, checkpoint, leaderEpoch) {
        let newCheckpoint = checkpoint;
        if (leaderEpoch !== newCheckpoint.epoch) {
            const lastSummaryState = await this.loadStateFromSummary(tenantId, documentId, gitManager, logger);
            if (lastSummaryState === undefined) {
                newCheckpoint.epoch = leaderEpoch;
            }
            else {
                // Log offset should never move backwards.
                const logOffset = newCheckpoint.logOffset;
                newCheckpoint = lastSummaryState;
                newCheckpoint.epoch = leaderEpoch;
                ++newCheckpoint.term;
                newCheckpoint.durableSequenceNumber = lastSummaryState.sequenceNumber;
                newCheckpoint.logOffset = logOffset;
                // Now create the summary.
                await this.createSummaryWithLatestTerm(gitManager, newCheckpoint, documentId);
                logger === null || logger === void 0 ? void 0 : logger.info(`Created a summary on epoch tick`, {
                    messageMetaData: {
                        documentId,
                        tenantId,
                    },
                });
            }
        }
        return newCheckpoint;
    }
    async createSummaryWithLatestTerm(gitManager, checkpoint, documentId) {
        const existingRef = await gitManager.getRef(encodeURIComponent(documentId));
        const [lastCommit, scribeContent] = await Promise.all([
            gitManager.getCommit(existingRef.object.sha),
            gitManager.getContent(existingRef.object.sha, ".serviceProtocol/scribe")
        ]);
        const scribe = (0, common_utils_1.toUtf8)(scribeContent.content, scribeContent.encoding);
        const serviceProtocolEntries = (0, protocol_base_1.generateServiceProtocolEntries)(JSON.stringify(checkpoint), scribe);
        const [serviceProtocolTree, lastSummaryTree] = await Promise.all([
            gitManager.createTree({ entries: serviceProtocolEntries }),
            gitManager.getTree(lastCommit.tree.sha, false),
        ]);
        const newTreeEntries = lastSummaryTree.tree
            .filter((value) => value.path !== ".serviceProtocol")
            .map((value) => {
            const createTreeEntry = {
                mode: value.mode,
                path: value.path,
                sha: value.sha,
                type: value.type,
            };
            return createTreeEntry;
        });
        newTreeEntries.push({
            mode: protocol_definitions_1.FileMode.Directory,
            path: ".serviceProtocol",
            sha: serviceProtocolTree.sha,
            type: "tree",
        });
        const gitTree = await gitManager.createGitTree({ tree: newTreeEntries });
        const commitParams = {
            author: {
                date: new Date().toISOString(),
                email: "praguertdev@microsoft.com",
                name: "Routerlicious Service",
            },
            message: `Term Change Summary @T${checkpoint.term}S${checkpoint.sequenceNumber}`,
            parents: [lastCommit.sha],
            tree: gitTree.sha,
        };
        // Finally commit the summary and update the ref.
        const commit = await gitManager.createCommit(commitParams);
        await gitManager.upsertRef(documentId, commit.sha);
    }
}
exports.DeliLambdaFactory = DeliLambdaFactory;
//# sourceMappingURL=lambdaFactory.js.map