"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.ScribeLambdaFactory = void 0;
const events_1 = require("events");
const util_1 = require("util");
const server_services_core_1 = require("@fluidframework/server-services-core");
const protocol_definitions_1 = require("@fluidframework/protocol-definitions");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const utils_1 = require("../utils");
const checkpointManager_1 = require("./checkpointManager");
const lambda_1 = require("./lambda");
const summaryReader_1 = require("./summaryReader");
const summaryWriter_1 = require("./summaryWriter");
const utils_2 = require("./utils");
const DefaultScribe = {
    lastClientSummaryHead: undefined,
    logOffset: -1,
    minimumSequenceNumber: 0,
    protocolState: {
        members: [],
        minimumSequenceNumber: 0,
        proposals: [],
        sequenceNumber: 0,
        values: [],
    },
    sequenceNumber: 0,
    lastSummarySequenceNumber: 0,
};
class ScribeLambdaFactory extends events_1.EventEmitter {
    constructor(mongoManager, documentCollection, messageCollection, producer, tenantManager, serviceConfiguration, enableWholeSummaryUpload) {
        super();
        this.mongoManager = mongoManager;
        this.documentCollection = documentCollection;
        this.messageCollection = messageCollection;
        this.producer = producer;
        this.tenantManager = tenantManager;
        this.serviceConfiguration = serviceConfiguration;
        this.enableWholeSummaryUpload = enableWholeSummaryUpload;
    }
    async create(config, context) {
        var _a, _b, _c, _d, _e, _f, _g, _h;
        let document;
        let gitManager;
        let lastCheckpoint;
        let summaryReader;
        let latestSummary;
        let opMessages;
        const { tenantId, documentId } = config;
        const messageMetaData = {
            documentId,
            tenantId,
        };
        const scribeSessionMetric = (0, utils_1.createSessionMetric)(tenantId, documentId, server_services_telemetry_1.LumberEventName.ScribeSessionResult, this.serviceConfiguration);
        try {
            const tenant = await this.tenantManager.getTenant(tenantId, documentId);
            gitManager = tenant.gitManager;
            summaryReader = new summaryReader_1.SummaryReader(tenantId, documentId, gitManager, this.enableWholeSummaryUpload);
            [latestSummary, document] = await Promise.all([
                summaryReader.readLastSummary(),
                this.documentCollection.findOne({ documentId, tenantId }),
            ]);
            if (!(0, utils_1.isDocumentValid)(document)) {
                // Document sessions can be joined (via Alfred) after a document is functionally deleted.
                // If the document doesn't exist or is marked for deletion then we trivially accept every message.
                (_a = context.log) === null || _a === void 0 ? void 0 : _a.error(`Received attempt to connect to a missing/deleted document.`, { messageMetaData });
                return new utils_1.NoOpLambda(context);
            }
            if (!(0, utils_1.isDocumentSessionValid)(document, this.serviceConfiguration)) {
                // Session for this document is either nonexistent or exists in a different location.
                const errMsg = `Received attempt to connect to invalid session: ${JSON.stringify(document.session)}`;
                (_b = context.log) === null || _b === void 0 ? void 0 : _b.error(errMsg, { messageMetaData });
                if (this.serviceConfiguration.enforceDiscoveryFlow) {
                    // This can/will prevent any users from creating a valid session in this location
                    // for the liftime of this NoOpLambda. This is not ideal; however, throwing an error
                    // to prevent lambda creation would mark the document as corrupted, which is worse.
                    return new utils_1.NoOpLambda(context);
                }
            }
            // Fetch pending ops from scribeDeltas collection
            const dbMessages = await this.messageCollection.find({ documentId, tenantId }, { "operation.sequenceNumber": 1 });
            opMessages = dbMessages.map((message) => message.operation);
        }
        catch (error) {
            (_c = context.log) === null || _c === void 0 ? void 0 : _c.error(`Scribe lambda creation failed. Exception: ${(0, util_1.inspect)(error)}`);
            await this.sendLambdaStartResult(tenantId, documentId, { lambdaName: server_services_core_1.LambdaName.Scribe, success: false });
            scribeSessionMetric === null || scribeSessionMetric === void 0 ? void 0 : scribeSessionMetric.error("Scribe lambda creation failed", error);
            throw error;
        }
        // Restore scribe state if not present in the cache. Mongodb casts undefined as null so we are checking
        // both to be safe. Empty sring denotes a cache that was cleared due to a service summary
        if (document.scribe === undefined || document.scribe === null) {
            (_d = context.log) === null || _d === void 0 ? void 0 : _d.info(`New document. Setting empty scribe checkpoint`, { messageMetaData });
            lastCheckpoint = DefaultScribe;
            opMessages = [];
        }
        else if (document.scribe === "") {
            (_e = context.log) === null || _e === void 0 ? void 0 : _e.info(`Existing document. Fetching checkpoint from summary`, { messageMetaData });
            if (!latestSummary.fromSummary) {
                (_f = context.log) === null || _f === void 0 ? void 0 : _f.error(`Summary can't be fetched`, { messageMetaData });
                lastCheckpoint = DefaultScribe;
                opMessages = [];
            }
            else {
                lastCheckpoint = JSON.parse(latestSummary.scribe);
                opMessages = latestSummary.messages;
                // Since the document was originated elsewhere or cache was cleared, logOffset info is irrelavant.
                // Currently the lambda checkpoints only after updating the logOffset so setting this to lower
                // is okay. Conceptually this is similar to default checkpoint where logOffset is -1. In this case,
                // the sequence number is 'n' rather than '0'.
                lastCheckpoint.logOffset = -1;
                (_g = context.log) === null || _g === void 0 ? void 0 : _g.info(JSON.stringify(lastCheckpoint));
            }
        }
        else {
            lastCheckpoint = JSON.parse(document.scribe);
            (_h = context.log) === null || _h === void 0 ? void 0 : _h.info(`Restoring checkpoint from db. Seq no: ${lastCheckpoint.sequenceNumber}`);
        }
        // Filter and keep ops after protocol state
        const opsSinceLastSummary = opMessages
            .filter((message) => message.sequenceNumber > lastCheckpoint.protocolState.sequenceNumber);
        let expectedSequenceNumber = lastCheckpoint.protocolState.sequenceNumber + 1;
        for (const message of opsSinceLastSummary) {
            if (message.sequenceNumber !== expectedSequenceNumber) {
                const error = new Error(`Invalid message sequence from checkpoint/summary.`
                    + `Current message @${message.sequenceNumber}.`
                    + `Expected message @${expectedSequenceNumber}`);
                scribeSessionMetric === null || scribeSessionMetric === void 0 ? void 0 : scribeSessionMetric.error("Invalid message sequence from checkpoint/summary", error);
                await this.sendLambdaStartResult(tenantId, documentId, {
                    lambdaName: server_services_core_1.LambdaName.Scribe,
                    success: false,
                });
                throw error;
            }
            ++expectedSequenceNumber;
        }
        const protocolHandler = (0, utils_2.initializeProtocol)(lastCheckpoint.protocolState, latestSummary.term);
        const lastSummaryMessages = latestSummary.messages;
        const summaryWriter = new summaryWriter_1.SummaryWriter(tenantId, documentId, gitManager, this.messageCollection, this.enableWholeSummaryUpload, lastSummaryMessages);
        const checkpointManager = new checkpointManager_1.CheckpointManager(context, tenantId, documentId, this.documentCollection, this.messageCollection);
        const scribeLambda = new lambda_1.ScribeLambda(context, document.tenantId, document.documentId, summaryWriter, summaryReader, undefined, checkpointManager, lastCheckpoint, this.serviceConfiguration, this.producer, protocolHandler, latestSummary.term, latestSummary.protocolHead, opsSinceLastSummary, scribeSessionMetric);
        await this.sendLambdaStartResult(tenantId, documentId, {
            lambdaName: server_services_core_1.LambdaName.Scribe,
            success: true,
        });
        return scribeLambda;
    }
    async dispose() {
        await this.mongoManager.close();
    }
    async sendLambdaStartResult(tenantId, documentId, contents) {
        const controlMessage = {
            type: server_services_core_1.ControlMessageType.LambdaStartResult,
            contents,
        };
        const operation = {
            clientSequenceNumber: -1,
            contents: null,
            data: JSON.stringify(controlMessage),
            referenceSequenceNumber: -1,
            traces: this.serviceConfiguration.enableTraces ? [] : undefined,
            type: protocol_definitions_1.MessageType.Control,
        };
        return (0, utils_2.sendToDeli)(tenantId, documentId, this.producer, operation);
    }
}
exports.ScribeLambdaFactory = ScribeLambdaFactory;
//# sourceMappingURL=lambdaFactory.js.map